import logging
log = logging.getLogger('onegov.user')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from translationstring import TranslationStringFactory
_ = TranslationStringFactory('onegov.user')  # noqa

from onegov.user.auth import Auth
from onegov.user.collection import UserCollection
from onegov.user.model import User
from onegov.user.utils import (
    is_valid_yubikey,
    is_valid_yubikey_format,
    yubikey_otp_to_serial
)

__all__ = [
    'Auth',
    'is_valid_yubikey',
    'is_valid_yubikey_format',
    'User',
    'UserCollection',
    'yubikey_otp_to_serial'
]



from onegov.core.crypto import hash_password, verify_password
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import JSON, UUID, LowercaseText
from onegov.core.utils import remove_repeated_spaces
from onegov.search import ORMSearchable
from onegov.user.utils import is_valid_yubikey_format
from onegov.user.utils import yubikey_otp_to_serial
from sqlalchemy import Boolean, Column, Index, Text, func
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import deferred
from uuid import uuid4


class User(Base, TimestampMixin, ORMSearchable):
    """ Defines a generic user. """

    __tablename__ = 'users'

    es_language = 'de'  # XXX add to database in the future
    es_properties = {
        'username': {'type': 'text'},
        'realname': {'type': 'text'},
        'userprofile': {'type': 'text'}
    }
    es_public = False

    @property
    def es_suggestion(self):
        return (self.realname or self.username, self.username)

    @property
    def userprofile(self):
        data = []

        if self.data:
            for value in self.data.values():
                if value and isinstance(value, str):
                    data.append(value)

        return data

    #: the user id is a uuid because that's more secure (no id guessing)
    id = Column(UUID, nullable=False, primary_key=True, default=uuid4)

    #: the username may be any string, but will usually be an email address
    username = Column(LowercaseText, unique=True, nullable=False)

    #: the password is stored with the hashing algorithm defined by onegov.core
    password_hash = Column(Text, nullable=False)

    #: the role is relevant for security in onegov.core
    role = Column(Text, nullable=False)

    #: the real name of the user for display (use the :attr:`name` property
    #: to automatically get the name or the username)
    realname = Column(Text, nullable=True)

    #: extra data that may be stored with this user, the format and content
    #: of this data is defined by the consumer of onegov.user
    #: by default, this data is only loaded by request, so if you need to
    #: load a lot of data columns, do something like this::
    #:
    #:     session.query(User).options(undefer("data"))
    #:
    data = deferred(Column(JSON, nullable=True))

    #: two-factor authentication schemes are enabled with this property
    #: if no two-factor auth is used, the value is NULL, if one *is* used,
    #: there should be a dictionary with the type of the two-factor
    #: authentication as well as custom values required by the two-factor
    #: implementation.
    #:
    #: e.g.::
    #:
    #:      {
    #:          'type': 'yubikey',
    #:          'data': 'ccccccbcgujh'
    #:      }
    #:
    #: Note that 'data' could also be a nested dictionary!
    #:
    second_factor = Column(JSON, nullable=True)

    #: true if the user is active
    active = Column(Boolean, nullable=False, default=True)

    __table_args__ = (
        Index('lowercase_username', func.lower(username), unique=True),
    )

    @hybrid_property
    def title(self):
        """ Returns the realname or the username of the user, depending on
        what's available first. """
        return self.realname or self.username

    @title.expression
    def title(cls):
        return func.coalesce(func.nullif(cls.realname, ''), cls.username)

    @hybrid_property
    def password(self):
        """ An alias for :attr:`password_hash`. """
        return self.password_hash

    @password.setter
    def password(self, value):
        """ When set, the given password in cleartext is hashed using
        onegov.core's default hashing algorithm.

        """
        self.password_hash = hash_password(value)

    def is_matching_password(self, password):
        """ Returns True if the given password (cleartext) matches the
        stored password hash.

        """
        return verify_password(password, self.password_hash)

    @property
    def initials(self):
        """ Takes the name and returns initials which are at most two
        characters wide.

        Examples:

        admin => A
        nathan.drake@example.org => ND
        Victor Sullivan => VS
        Charles Montgomery Burns => CB

        """

        # for e-mail addresses assume the dot splits the name and use
        # the first two parts of said split (probably won't have a middle
        # name in the e-mail address)
        if not self.realname:
            username = self.username.split('@')[0]
            parts = username.split('.')[:2]

        # for real names split by space and assume that with more than one
        # part that the first and last part are the most important to get rid
        # of middlenames
        else:
            parts = remove_repeated_spaces(self.realname).split(' ')

            if len(parts) > 2:
                parts = (parts[0], parts[-1])

        return ''.join(p[0] for p in parts).upper()

    @property
    def has_yubikey(self):
        if not self.second_factor:
            return False

        return self.second_factor.get('type') == 'yubikey'

    @property
    def yubikey(self):
        if not self.has_yubikey:
            return None

        return self.second_factor.get('data')

    @yubikey.setter
    def yubikey(self, yubikey):
        if not yubikey:
            self.second_factor = None
        else:
            assert is_valid_yubikey_format(yubikey)

            self.second_factor = {
                'type': 'yubikey',
                'data': yubikey[:12]
            }

    @property
    def yubikey_serial(self):
        """ Returns the yubikey serial of the yubikey associated with this
        user (if any).

        """

        yubikey = self.yubikey

        return yubikey and yubikey_otp_to_serial(yubikey) or None



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from collections import defaultdict
from onegov.core.upgrade import upgrade_task
from onegov.core.orm.types import JSON
from onegov.user import User, UserCollection
from sqlalchemy import Boolean, Column, Text
from sqlalchemy.sql import text


@upgrade_task('Add second_factor column')
def add_second_factor_column(context):
    context.operations.add_column(
        'users', Column('second_factor', JSON, nullable=True)
    )


@upgrade_task('Add active column')
def add_active_column(context):
    context.operations.add_column(
        'users', Column('active', Boolean, nullable=True, default=True)
    )

    for user in context.session.query(User).all():
        user.active = True

    context.session.flush()
    context.operations.alter_column('users', 'active', nullable=False)


@upgrade_task('Add realname column')
def add_realname_column(context):
    context.operations.add_column(
        'users', Column('realname', Text, nullable=True))

    for user in context.session.query(User).all():
        user.realname = (user.data and user.data or {}).get('name')

    context.session.flush()


def change_ownership_by_name(context, old_username, new_username):
    # transfer all ownership without using models (which might or
    # might not be available here)
    if context.has_table('activities'):
        context.operations.execute("""
            UPDATE activities
            SET username = '{new_username}'
            WHERE username = '{old_username}'
        """.format(new_username=new_username, old_username=old_username))

    if context.has_table('attendees'):
        context.operations.execute("""
            UPDATE attendees
            SET username = '{new_username}'
            WHERE username = '{old_username}'
        """.format(new_username=new_username, old_username=old_username))

    if context.has_table('bookings'):
        context.operations.execute("""
            UPDATE bookings
            SET username = '{new_username}'
            WHERE username = '{old_username}'
        """.format(new_username=new_username, old_username=old_username))

    if context.has_table('invoice_items'):
        context.operations.execute("""
            UPDATE invoice_items
            SET username = '{new_username}'
            WHERE username = '{old_username}'
        """.format(new_username=new_username, old_username=old_username))


def change_ownership_by_id(context, old_userid, new_userid):
    if context.has_table('tickets'):
        context.operations.execute("""
            UPDATE tickets
            SET user_id = '{new_userid}'
            WHERE user_id = '{old_userid}'
        """.format(old_userid=old_userid.hex, new_userid=new_userid.hex))


@upgrade_task('Force lowercase usernames')
def force_lowercase_usernames(context):
    users = defaultdict(list)

    for user in context.session.query(User).all():
        users[user.username.lower()].append(user)

    temp_user = UserCollection(context.session).add(
        username='temp',
        password='temp',
        active=False,
        role='member',
    )

    for username, users in users.items():

        # simply change usernames that don't conflict with others
        if len(users) == 1:
            with context.session.no_autoflush:
                change_ownership_by_name(
                    context, users[0].username, 'temp')

                users[0].username = users[0].username.lower()
                context.session.flush()

                change_ownership_by_name(
                    context, 'temp', users[0].username)

            continue

        # from others select one user, move ownership to it and then
        # remove the other users -> that means that some of our users need
        # to reset their password...

        # the remaining user is selected by activity, last-change and role
        # we prefer active users over low-ranking users over less recently
        # updated users
        role_hierarchy = [
            'member',
            'editor',
            'admin'
        ]

        def sort_key(user):
            return (
                user.active,
                role_hierarchy.index(user.role),
                user.last_change,
            )

        remaining = sorted(users, key=sort_key)[-1]
        remaining_data = remaining.data or {}

        others = [u for u in users if u.id != remaining.id]

        for other in others:

            # keep as much user data as possible
            for key, value in (other.data or {}).items():
                if value and not remaining.data.get(key):
                    remaining_data[key] = value

            # change the userids
            change_ownership_by_id(context, other.id, remaining.id)

            # change the username
            change_ownership_by_name(context, other.username, 'temp')

            # delete the other user
            context.operations.execute("""
                DELETE from users
                WHERE id = '{}'
            """.format(other.id.hex))

        # switch the remaining user
        change_ownership_by_name(context, remaining.username, 'temp')

        remaining.username = remaining.username.lower()
        remaining.data = remaining_data
        context.session.flush()

        change_ownership_by_name(context, 'temp', remaining.username)

    # enforce the lowercase rule
    context.operations.create_index(
        'lowercase_username', 'users', [
            text('lower("username")')
        ], unique=True
    )

    # remove the temporary user
    context.session.delete(temp_user)
